/*----------------------------------------------------------------------
  File    : wlvq.c
  Contents: learning vector quantization visualization
  Author  : Christian Borgelt
  History : 2000.09.06 file created from wbcview.c
            2001.06.27 bug in function fselect removed
            2003.04.10 DI_LREXP changed to DI_DECAY
            2007.02.18 marker size added, command line eval. improved
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <windows.h>
#include <commdlg.h>
#include "wlvq.rh"
#include "wshades.h"
#include "lvqd.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "wlvq"      /* main window title */

#ifndef PATH_MAX
#define PATH_MAX    1024        /* max. len. of a path name */
#endif

/* --- abbreviations --- */
#define HINST       HINSTANCE
#define MKIRSC      MAKEINTRESOURCE
#define SDIMSG      SendDlgItemMessage
#define WINSTYLE    WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN

/* --- extensions --- */
#define MINXEXT     160         /* minimal x-extension of window */
#define MINYEXT     120         /* minimal y-extension of window */
#define INITXEXT    320         /* initial horizontal extension */
#define INITYEXT    320         /* initial vertical   extension */

/* --- edit box functions --- */
#define edt_set(h,i,s)    SetDlgItemText(h, i, s)
#define edt_get(h,i,s,n)  GetDlgItemText(h, i, s, n)
#define int_set(h,i,n)    SetDlgItemInt(h, i, n, 0)
#define int_get(h,i)      GetDlgItemInt(h, i, NULL, 0)

/* --- check box functions --- */
#define chk_set(h,i,s)    CheckDlgButton(h, i, s)
#define chk_get(h,i)      IsDlgButtonChecked(h, i)

/* --- combo box functions --- */
#define cmb_add(h,i,s)    SDIMSG(h, i, CB_ADDSTRING,    0, (LPARAM)(s))
#define cmb_rem(h,i,k)    SDIMSG(h, i, CB_DELETESTRING, k, 0)
#define cmb_set(h,i,k)    SDIMSG(h, i, CB_SETCURSEL,    k, 0)
#define cmb_get(h,i)      SDIMSG(h, i, CB_GETCURSEL,    0, 0)

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_DATA       (-8)       /* no data loaded */
#define E_COLOR      (-9)       /* cannot allocate colors */
#define E_UNKNOWN   (-21)       /* unkown error */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- coordinate transformation --- */
  double xoff, yoff;            /* offsets */
  double xscl, yscl;            /* scaling factors */
} TRANSFORM;                    /* (coordinate transformation) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
/* --- error messages --- */
static const char *errmsgs[] ={ /* error message texts */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file:\n%s",
  /* E_FREAD    -3 */  "read error on file:\n%s",
  /* E_FWRITE   -4 */  "write error on file:\n%s",
  /* E_METCNT   -5 */  "no numeric attribute",
  /* E_REFVEC   -6 */  "no reference vectors",
  /* E_RVINC    -7 */  "reference vectors are incompatible\n"
                         "with the loaded data",
  /* E_DATA     -8 */  "no data loaded",
  /* E_COLOR    -9 */  "cannot allocate colors",
  /*    -10 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_VALUE   -16 */  "file %s, record %d:\n"
                         "invalid value %s in field %d",
  /* E_FLDCNT  -17 */  "file %s, record %d:\n"
                         "%s%d field(s) instead of %d",
  /* E_EMPFLD  -18 */  "file %s, record %d:\n"
                         "empty name%s in field %d",
  /* E_DUPFLD  -19 */  "file %s, record %d:\n"
                         "duplicate field name %s",
  /* E_MISFLD  -20 */  "file %s, record %d:\n"
                         "missing field %s",
  /* E_UNKNOWN -21 */  "unknown error" };

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static HINST  hinst    = NULL;  /* program instance handle */
static HWND   hmain    = NULL;  /* handle of main window */
static HDC    hdc      = NULL;  /* handle of device context */
static char   buf[PATH_MAX +1024];  /* buffer for various purposes */

/* --- color information --- */
static SHADES *shades  = NULL;  /* color shades, brushes, and pens */
static HBRUSH *brushes = NULL;  /* for drawing the data points */
static HPEN   *pens    = NULL;  /* and the reference vectors */
static int    mrksize  = 5;     /* marker size */

static struct {                 /* -- color information -- */
  int    bkgnd;                 /* background color (black or white) */
  double coloff;                /* color offset (in degrees) */
  int    colcnt;                /* number of colors */
  int    shdcnt;                /* number of shades */
} colinfo = { SHD_WHITE, 0.0, 0, 5 };

/* --- brushes and colors --- */
static HPEN   p_black;          /* black lines */
static HPEN   p_white;          /* white lines */
static HBRUSH br_white;         /* white background */
static HBRUSH br_black;         /* black background */
static HBRUSH br_grey;          /* grey default brush */
static HBRUSH br_null;          /* hollow brush */

static int    running = 0;      /* whether training is running */

/*----------------------------------------------------------------------
  Error Reporting Function
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* error message */
  int        n;                 /* length of the formatted message */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs[-code];         /* get error message data */
  if (!msg) msg = errmsgs[-E_UNKNOWN];
  va_start(args, code);         /* get list of variable arguments */
  vsprintf(buf, msg, args);     /* get and format message */
  va_end(args);                 /* end variable argument evaluation */
  n = (int)strlen(buf) +1;      /* load alert box title */
  LoadString(hinst, CS_ALERT, buf +n, sizeof(buf) -n);
  MessageBox(hmain, buf, buf+n, MB_OK|MB_ICONERROR);
}  /* error() */                /* show alert box */

/*----------------------------------------------------------------------
  Color Functions
----------------------------------------------------------------------*/

static void col_delete (void)
{                               /* --- delete color shades */
  int i;                        /* loop variable */

  if (shades)  {                /* delete all shades/colors */
    shd_delete(shades); shades = NULL; }
  if (brushes) {                /* delete all brushes */
    for (i = colinfo.colcnt; --i >= 0; ) DeleteObject(brushes[i]);
    free(brushes); brushes = NULL;
  }
  if (pens) {                   /* delete all pens */
    for (i = colinfo.colcnt; --i >= 0; ) DeleteObject(pens[i]);
    free(pens);    pens    = NULL;
  }
}  /* col_delete() */

/*--------------------------------------------------------------------*/

static int col_create (void)
{                               /* --- get color shades */
  int i, k, s;                  /* loop variable, buffers */

  col_delete();                 /* delete existing shades/colors */
  if (!table) return 0;         /* check for existing data */
  colinfo.colcnt = (selatt.c_att < 0) ? 1
                 : att_valcnt(as_att(attset, map_nom[selatt.c_att]));
                                /* get the number of colors */
  k = (colinfo.shdcnt = (colinfo.bkgnd == SHD_BLACK) ? 3 : 5) -1;
  shades = shd_create(colinfo.coloff, colinfo.colcnt,
                      colinfo.bkgnd,  colinfo.shdcnt);
  if (!shades)  return -1;      /* create color shades */
  brushes = (HBRUSH*)malloc(colinfo.colcnt *sizeof(HBRUSH));
  if (!brushes) return -1;      /* create brushes for drawing */
  for (i = colinfo.colcnt; --i >= 0; )
    brushes[i] = CreateSolidBrush(shd_pixel(shades, i, k));
  pens    = (HPEN*) malloc(colinfo.colcnt *2 *sizeof(HPEN));
  if (!pens)    return -1;      /* create pens for drawing */
  s = mrksize >> 2; if (s < 1) s = 1;
  for (i = colinfo.colcnt; --i >= 0; ) {
    pens[i+i  ] = CreatePen(PS_SOLID, s, shd_pixel(shades, i, k));
    pens[i+i+1] = CreatePen(PS_SOLID, s, shd_pixel(shades, i, 1));
  }                             /* pens with light and dark color */
  return 0;                     /* return 'ok' */
}  /* col_create() */

/*----------------------------------------------------------------------
  Drawing Functions
----------------------------------------------------------------------*/

static void draw_vecs (HDC hdc, REFVEC *vecs, TRANSFORM *tf, int light)
{                               /* --- draw reference vectors */
  int i, k;                     /* loop variable, buffer */
  int s, o;                     /* size and offset for drawing */
  int x, y;                     /* pixel coordinates */

  assert(vecs && tf);           /* check the function arguments */
  assert(shadesb && pens);      /* check for existing shades */
  s = mrksize +2; o = s >> 1;   /* compute the marker size parameters */
  SelectObject(hdc, br_null);   /* get hollow brush for interior */
  for (i = rvcnt; --i >= 0; ) { /* traverse the reference vectors */
    x = (int)((vecs[i].x -tf->xoff) *tf->xscl +0.5) -o;
    y = (int)((tf->yoff -vecs[i].y) *tf->yscl +0.5) -o;
    k = rvcls[i];               /* get coordinates and class */
    SelectObject(hdc, pens[k +k +light]);
    Ellipse(hdc, x, y, x+s, y+s);
  }                             /* draw a hollow circle */
}  /* draw_vecs() */            /* for each reference vector */

/*--------------------------------------------------------------------*/

static void redraw (HDC hdc)
{                               /* --- redraw window contents */
  int       i, k;               /* loop variable, buffer */
  int       s, o;               /* size and offset for drawing */
  float     f;                  /* temporary buffer */
  TUPLE     *tpl;               /* to traverse the data tuples */
  int       xid, yid, clsid;    /* attribute identifiers */
  int       xtype, ytype;       /* attribute types */
  int       cls;                /* current class value */
  double    rx, ry;             /* real  coordinates */
  int       x,  y;              /* pixel coordinates */
  TRANSFORM tf;                 /* coordinate transformation */
  RECT      rc;                 /* clipping rectangle */
  HBRUSH    brush;              /* brush for drawing data points */

  /* --- draw background --- */
  GetClientRect(hmain, &rc);    /* get the window's client rectangle */
  FillRect(hdc, &rc, (colinfo.bkgnd == SHD_BLACK) ? br_black:br_white);
  if (!table || !shades || !brushes || !pens)
    return;                     /* check prerequisites for drawing */

  /* --- build transformation --- */
  tf.xoff = ranges[selatt.h_att].min; /* get coordinate offsets and  */
  tf.yoff = ranges[selatt.v_att].max; /* compute the scaling factors */
  tf.xscl = (rc.right -rc.left) /(ranges[selatt.h_att].max -tf.xoff);
  tf.yscl = (rc.bottom -rc.top) /(tf.yoff -ranges[selatt.v_att].min);

  /* --- draw data --- */
  clsid = (selatt.c_att < 0)    /* get class attribute identifier */
        ? -1 : map_nom[selatt.c_att];
  xid   = map_met[selatt.h_att];/* get numeric attribute ids. */
  yid   = map_met[selatt.v_att];/* and attribute types */
  xtype = att_type(as_att(attset, xid));
  ytype = att_type(as_att(attset, yid));
  s = mrksize +2; o = s >> 1;   /* compute the marker size parameters */
  SelectObject(hdc, (colinfo.bkgnd == SHD_BLACK) ? p_white : p_black);
  for (k = tab_tplcnt(table); --k >= 0; ) {
    tpl = tab_tpl(table, k);    /* traverse the data tuples */
    if (xtype == AT_INT) {      /* if x-coordinate is integer valued */
      i = tpl_colval(tpl, xid)->i;
      if (isnull(i)) continue;
      rx = (double)i; }         /* get and check the attribute value */
    else {                      /* if x-coordinate is real valued */
      f = tpl_colval(tpl, xid)->f;
      if (isnan(f))  continue;
      rx = (double)f;           /* get and check the attribute value */
    }
    if (ytype == AT_INT) {      /* if y-coordinate is integer valued */
      i = tpl_colval(tpl, yid)->i;
      if (isnull(i)) continue;
      ry = (double)i; }         /* get and check the attribute value */
    else {                      /* if y-coordinate is real valued */
      f = tpl_colval(tpl, yid)->f;
      if (isnan(f))  continue;
      ry = (double)f;           /* get and check the attribute value */
    }                           /* then compute the screen coords. */
    x     = (int)((rx -tf.xoff) *tf.xscl +0.5) -o;
    y     = (int)((tf.yoff -ry) *tf.yscl +0.5) -o;
    cls   = (clsid < 0) ?  -1  : tpl_colval(tpl, clsid)->i;
    brush = (cls < 0) ? br_grey : brushes[cls];
    SelectObject(hdc, brush);   /* get and set the brush for drawing */
    Ellipse(hdc, x, y, x+s, y+s);
  }                             /* draw a circle at the data point */

  /* --- draw reference vectors --- */
  if (hist) {                   /* if there is a vector history */
    for (k = 0; k < histcnt; k++)
      draw_vecs(hdc, hist[k], &tf, 1);
  }                             /* draw reference vectors */
  if (rvecs) draw_vecs(hdc, rvecs, &tf, 0);

  sprintf(buf, "%s %10ld", PRGNAME, histcnt);
  SetWindowText(hmain, buf);    /* show the current epoch */
}  /* redraw() */

/*----------------------------------------------------------------------
  Neural Network Training Function
----------------------------------------------------------------------*/

static void CALLBACK train (HWND hwnd, UINT msg, UINT id, DWORD systime)
{                               /* --- do one training step */
  RECT      rc;                 /* window rectangle */
  TRANSFORM tf;                 /* coordinate transformation */

  if (!running) return;         /* check running flag */
  lvq_train();                  /* do one training step */
  GetClientRect(hmain, &rc);    /* get the window's client rectangle */
  tf.xoff = ranges[selatt.h_att].min; /* get coordinate offsets and  */
  tf.yoff = ranges[selatt.v_att].max; /* compute the scaling factors */
  tf.xscl = (rc.right -rc.left) /(ranges[selatt.h_att].max -tf.xoff);
  tf.yscl = (rc.bottom -rc.top) /(tf.yoff -ranges[selatt.v_att].min);
  draw_vecs(hdc, hist[histcnt-1], &tf, 1);
  draw_vecs(hdc, rvecs,           &tf, 0);
                                /* (re)draw the reference vectors */
  sprintf(buf, "%s %10ld", PRGNAME, histcnt);
  SetWindowText(hmain, buf);    /* show the current epoch */
  SetTimer(hwnd, 1, (UINT)(lvqpar.delay *1000), (TIMERPROC)train);
}  /* train() */                /* reinstall timer */

/*----------------------------------------------------------------------
  Dialog Functions
----------------------------------------------------------------------*/

static const char* fselect (char *title, int load)
{                               /* --- select a file name */
  static char fname[_MAX_PATH+1] = "";  /* file name */
  static char path [_MAX_PATH+1];       /* and path */
  static char filter[] =        /* file selection filter */
    "all files (*.*)\0*.*\0";   /* (filter for all files) */
  static OPENFILENAME ofn = {   /* file selector structure */
    sizeof(OPENFILENAME),       /* lStructSize */
    NULL,   NULL,               /* hwndOwner, hInstance */
    filter, NULL,               /* lpstrFilter, lpstrCustomFilter */
    0, 1,                       /* nMaxCustFilter, nFilterIndex */
    fname, _MAX_PATH,           /* lpstrFile, nMaxFile */
    NULL, _MAX_FNAME+_MAX_EXT,  /* lpstrFileTitle, nMaxFileTitle */
    path,                       /* lpstrInitialDir */
    "Select File...",           /* lpstrTitle (with default init.) */
    OFN_PATHMUSTEXIST, 0, 0,    /* Flags, nFileOffset, nFileExtension */
    NULL, 0L,                   /* lpstrDefExt, lCustData */
    NULL, NULL };               /* lpfnHook, lpTemplateName */

  ofn.hwndOwner  = hmain;        /* set parent window handle */
  ofn.lpstrTitle = title;       /* and dialog box title */
  ofn.Flags = (load) ? OFN_FILEMUSTEXIST : OFN_PATHMUSTEXIST;
  if (GetOpenFileName(&ofn) == 0)
    return NULL;                /* open file selector box */
  return fname;                 /* return name of selected file */
}  /* fselect() */

/*--------------------------------------------------------------------*/

static void mn_enable(int enable)
{                               /* --- enable/disable menu items */
  HMENU        menu;            /* handle of window menu */
  MENUITEMINFO mii;             /* for menu customization */
  UINT         flag;            /* enable/disable flag */

  menu = GetMenu(hmain);        /* get menu handle */
  flag = (enable) ? MF_ENABLED : MF_GRAYED;
  EnableMenuItem(menu, MI_DATA, flag);    /* enable/disable context- */
  EnableMenuItem(menu, MI_LOAD, flag);    /* sensitive menu items */
  EnableMenuItem(menu, MI_ATTS, flag);
  mii.cbSize     = sizeof(MENUITEMINFO);  /* set either the */
  mii.fMask      = MIIM_ID|MIIM_TYPE;     /* 'Start Training' or */
  mii.fType      = MFT_STRING;            /* 'Stop Training' item */
  mii.wID        =  (enable) ? MI_START : MI_STOP;
  LoadString(hinst, (enable) ? CS_START : CS_STOP, buf, 80);
  mii.dwTypeData = buf; mii.cch = (UINT)strlen(buf);
  SetMenuItemInfo(menu, (enable) ? MI_STOP : MI_START, FALSE, &mii);
}  /* mn_enable() */

/*----------------------------------------------------------------------
  Dialog Box Procedures
----------------------------------------------------------------------*/

BOOL CALLBACK dp_format (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- characters dialog box proc. */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      chk_set(hwnd, DI_NAMES,   datafmt.names);
      edt_set(hwnd, DI_RECORD,  datafmt.recseps);
      edt_set(hwnd, DI_FIELD,   datafmt.fldseps);
      edt_set(hwnd, DI_BLANK,   datafmt.blanks);
      edt_set(hwnd, DI_NULL,    datafmt.nullchs);
      edt_set(hwnd, DI_COMMENT, datafmt.comment);
      return 1;                 /* set defaults from parameter block */

    case WM_COMMAND:            /* command (button pressed etc.) */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          datafmt.names = chk_get(hwnd, DI_NAMES) ? 1 : 0;
          edt_get(hwnd, DI_RECORD,  datafmt.recseps, FMTCHRLEN);
          edt_get(hwnd, DI_FIELD,   datafmt.fldseps, FMTCHRLEN);
          edt_get(hwnd, DI_BLANK,   datafmt.blanks,  FMTCHRLEN);
          edt_get(hwnd, DI_NULL,    datafmt.nullchs, FMTCHRLEN);
          edt_get(hwnd, DI_COMMENT, datafmt.comment, FMTCHRLEN);
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;

    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_format() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_params (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- paremeters dialog box proc. */
  int i;                        /* loop variable, buffer */

  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      for (i = CS_CENTER; i <= CS_DATAPT; i++) {
        LoadString(hinst, i, buf, sizeof(buf));
        cmb_add(hwnd, DI_IMODE, buf);
      }                         /* fill vector init. list */
      cmb_set(hwnd, DI_IMODE, lvqpar.imode);
      sprintf(buf, "%d", lvqpar.vpcls); edt_set(hwnd, DI_VPCLS, buf);
      sprintf(buf, "%g", lvqpar.roff);  edt_set(hwnd, DI_ROFF,  buf);
      sprintf(buf, "%g", lvqpar.lrate); edt_set(hwnd, DI_LRATE, buf);
      sprintf(buf, "%g", lvqpar.decay); edt_set(hwnd, DI_DECAY, buf);
      sprintf(buf, "%g", lvqpar.delay); edt_set(hwnd, DI_DELAY, buf);
      return 1;                 /* set defaults from parameter block */

    case WM_COMMAND:            /* command (button pressed etc.) */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          lvqpar.imode = (int)cmb_get(hwnd, DI_IMODE);
                                /* vector initialization mode */
          edt_get(hwnd, DI_VPCLS, buf, 80);
          lvqpar.vpcls = atoi(buf);
          if      (lvqpar.vpcls < 1)     lvqpar.vpcls = 1;
          else if (lvqpar.vpcls > 100)   lvqpar.vpcls = 100;
                                /* number of vectors per class */
          edt_get(hwnd, DI_ROFF, buf, 80);
          lvqpar.roff  = atof(buf);
          if      (lvqpar.roff  < 0)     lvqpar.roff  = 0;
          else if (lvqpar.roff  > 100)   lvqpar.roff  = 10;
                                /* random offset range */
          edt_get(hwnd, DI_LRATE, buf, 80);
          lvqpar.lrate = atof(buf);
          if      (lvqpar.lrate < -100)  lvqpar.lrate = -100;
          else if (lvqpar.lrate > 10)    lvqpar.lrate = 10;
                                /* initial learning rate */
          edt_get(hwnd, DI_DECAY, buf, 80);
          lvqpar.decay = atof(buf);
          if      (lvqpar.decay > 1)     lvqpar.decay = 1;
                                /* learning rate decay exponent */
          edt_get(hwnd, DI_DELAY, buf, 80);
          lvqpar.delay = atof(buf);
          if      (lvqpar.delay < 0.001) lvqpar.delay = 0.001;
          else if (lvqpar.delay > 10)    lvqpar.delay = 10;
                                /* delay between epochs */
          EndDialog(hwnd, 0);   /* terminate the dialog */
          return 1;
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;

    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* return 'not handled' */
}  /* db_params() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_atts (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- attributes dialog box proc. */
  int i;                        /* loop variable, buffer */

  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      if (nms_met) {            /* if there are numeric attributes */
        for (i = 0; i < metcnt; i++)
          cmb_add(hwnd, DI_H_ATT, nms_met[i]);
        cmb_set(hwnd, DI_H_ATT, selatt.h_att);
      }                         /* add them to the list box */
      sprintf(buf, "%g", selatt.h_rng.min);   /* set minimal value */
      edt_set(hwnd, DI_H_MIN, buf);    /* for horizontal direction */
      sprintf(buf, "%g", selatt.h_rng.max);   /* set maximal value */
      edt_set(hwnd, DI_H_MAX, buf);    /* for horizontal direction */
      if (nms_met) {            /* if there are numeric attributes */
        for (i = 0; i < metcnt; i++)
          cmb_add(hwnd, DI_V_ATT, nms_met[i]);
        cmb_set(hwnd, DI_V_ATT, selatt.v_att);
      }                         /* add them to the list box */
      sprintf(buf, "%g", selatt.v_rng.min);   /* set minimal value */
      edt_set(hwnd, DI_V_MIN, buf);      /* for vertical direction */
      sprintf(buf, "%g", selatt.v_rng.max);   /* set maximal value */
      edt_set(hwnd, DI_V_MAX, buf);      /* for vertical direction */
      if (nms_nom) {            /* if there are symbolic attribs. */
        LoadString(hinst, CS_NONE, buf, sizeof(buf));
        cmb_add(hwnd, DI_C_ATT, buf);
        for (i = 0; i < nomcnt; i++)
          cmb_add(hwnd, DI_C_ATT, nms_nom[i]);
        cmb_set(hwnd, DI_C_ATT, selatt.c_att +1);
      }                         /* add them to the list box */
      return 1;                 /* set defaults from parameter block */

    case WM_COMMAND:            /* command (button pressed etc.) */
      switch (LOWORD(wp)) {     /* evaluate command code */

        case DI_H_ATT:          /* if attribute for horizontal dir. */
          if ((HIWORD(wp) != CBN_SELCHANGE) || !ranges)
            break;              /* ignore anything but sel. change */
          i = (int)cmb_get(hwnd, DI_H_ATT);   /* get new selection */
          sprintf(buf, "%g", ranges[i].min);  /* set minimal value */
          edt_set(hwnd, DI_H_MIN, buf);       /* for horizontal dir. */
          sprintf(buf, "%g", ranges[i].max);  /* set maximal value */
          edt_set(hwnd, DI_H_MAX, buf);       /* for horizontal dir. */
          break;

        case DI_V_ATT:          /* if attribute for vertical dir. */
          if ((HIWORD(wp) != CBN_SELCHANGE) || !ranges)
            break;              /* ignore anything but sel. change */
          i = (int)cmb_get(hwnd, DI_V_ATT);   /* get new selection */
          sprintf(buf, "%g", ranges[i].min);  /* set minimal value */
          edt_set(hwnd, DI_V_MIN, buf);       /* for horizontal dir. */
          sprintf(buf, "%g", ranges[i].max);  /* set maximal value */
          edt_set(hwnd, DI_V_MAX, buf);       /* for horizontal dir. */
          break;

        case IDOK:              /* if 'Ok' button pressed */
          selatt.h_att = (int)cmb_get(hwnd, DI_H_ATT);
          edt_get(hwnd, DI_H_MIN, buf, 80);   /* get min. horz. value */
          ranges[selatt.h_att].min = selatt.h_rng.min = atof(buf);
          edt_get(hwnd, DI_H_MAX, buf, 80);   /* get max. horz. value */
          ranges[selatt.h_att].max = selatt.h_rng.max = atof(buf);
          selatt.v_att = (int)cmb_get(hwnd, DI_V_ATT);
          edt_get(hwnd, DI_V_MIN, buf, 80);   /* get min. vert. value */
          ranges[selatt.v_att].min = selatt.v_rng.min = atof(buf);
          edt_get(hwnd, DI_V_MAX, buf, 80);   /* get max. vert. value */
          ranges[selatt.v_att].max = selatt.v_rng.max = atof(buf);
          selatt.c_att = (int)cmb_get(hwnd, DI_C_ATT) -1;
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* return 'ok' */

        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;

    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_atts() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_colors (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- colors dialog box procedure */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* dialog initialization */
      LoadString(hinst, CS_BLACK,  buf, sizeof(buf));
      cmb_add(hwnd, DI_BGCOLOR, buf);
      LoadString(hinst, CS_WHITE,  buf, sizeof(buf));
      cmb_add(hwnd, DI_BGCOLOR, buf);   /* set background color */
      cmb_set(hwnd, DI_BGCOLOR, (colinfo.bkgnd == SHD_BLACK) ? 0 : 1);
      sprintf(buf, "%g", colinfo.coloff);
      edt_set(hwnd, DI_COLOFF, buf);    /* set color offset */
      sprintf(buf, "%d", mrksize);
      edt_set(hwnd, DI_SIZE, buf);      /* set marker size */
      return 1;

    case WM_COMMAND:            /* command (button pressed etc.) */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          edt_get(hwnd, DI_SIZE, buf, 80);
          mrksize = atoi(buf);  /* get the marker size */
          if (mrksize < 1) mrksize = 1;
          colinfo.bkgnd = (cmb_get(hwnd, DI_BGCOLOR) == 0)
                        ? SHD_BLACK : SHD_WHITE;
          edt_get(hwnd, DI_COLOFF, buf, 80);
          colinfo.coloff = atof(buf);
          if (col_create() != 0)/* get background color */
            error(E_COLOR);     /* and color offset, */
          EndDialog(hwnd, 0);   /* create new color shades, */
          return 1;             /* and terminate the dialog */
        case IDCANCEL:          /* if 'Cancel' button pressed, */
          EndDialog(hwnd, -1); return 1;   /* terminate dialog */
      } break;

    case WM_CLOSE:              /* dialog box has been closed */
      EndDialog(hwnd, -1); return 1;      /* terminate dialog */
  }
  return 0;                     /* everything else not handled */
}  /* dp_colors() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_about (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'about' dialog box procedure */
  switch (msg) {                /* evaluate message */
    case WM_COMMAND:            /* command (button pressed etc.) */
      if (wp != IDOK) break;    /* process only 'Ok' button */
    case WM_CLOSE:              /* dialog box closed */
      EndDialog(hwnd, 0); return 1;
  }                             /* terminate dialog */
  return 0;                     /* everything else not handled */
}  /* dp_about() */

/*----------------------------------------------------------------------
  Window Procedures
----------------------------------------------------------------------*/

LRESULT CALLBACK wp_main (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- main window procedure */
  static PAINTSTRUCT ps;        /* for BeginPaint call */
  static int  szmv = 0;         /* size/move flag */
  const  char *fname;           /* name of selected file */
  HDC    hdc;                   /* handle of device context */
  int    r;                     /* result data load function */

  switch (msg) {                /* evaluate message */
    case WM_GETMINMAXINFO:      /* window is about to be resized */
      ((LPMINMAXINFO)lp)->ptMinTrackSize.x = MINXEXT;
      ((LPMINMAXINFO)lp)->ptMinTrackSize.y = MINYEXT;
      break;                    /* set a minimal width and height */

    case WM_ENTERSIZEMOVE:      /* if window is resized or moved, */
      szmv = 1; break;          /* inhibit redrawing with a flag */
    case WM_SIZING:             /* if window is actually resized, */
      szmv = 2; break;          /* note this in the flag */
    case WM_EXITSIZEMOVE:       /* if resizing/moving is finished */
      if (szmv > 1) InvalidateRect(hwnd, NULL, 0);
      szmv = 0; break;          /* redraw the window contents */

    case WM_PAINT:              /* --- if to paint window */
      if (szmv) break;          /* during resize do not redraw */
      hdc = BeginPaint(hwnd, &ps);
      if (!hdc) return 0;       /* get device context for drawing */
      redraw(hdc);              /* redraw the window contents */
      EndPaint(hwnd, &ps);      /* release the device context */
      break;

    case WM_COMMAND:            /* --- if menu item selected */
      switch (wp) {             /* evaluate item */
        /* --- Menu File --- */
        case MI_DATA:           /* File > Load Data... */
          LoadString(hinst, CS_DATA, buf, sizeof(buf));
          fname = fselect(buf, 1);
          if (!fname) return -1;/* let the user select a file */
          r = tv_load(fname, 0.1); /* load a data table */
          if (r != 0) {         /* check for a read error */
            error(E_FREAD, fname); return 0; }
          if (col_create() != 0) { error(E_COLOR); return -1; }
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* redraw the window contents */

        case MI_FORMAT:         /* File > Data Format... */
          DialogBox(hinst, MKIRSC(DB_FORMAT), hwnd, (DLGPROC)dp_format);
          return 0;             /* open the dialog box */

        case MI_LOAD:           /* File > Load Vectors... */
          LoadString(hinst, CS_LOAD, buf, sizeof(buf));
          fname = fselect(buf, 1);
          if (!fname) return -1;/* let the user select a file */
          r = lvq_load(fname);  /* load reference vectors */
          if (r != 0) { error(r, fname); return -1; }
          if (col_create() != 0) { error(E_COLOR); return -1; }
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* redraw the window contents */

        case MI_SAVE:           /* File > Save Vectors... */
          LoadString(hinst, CS_SAVE, buf, sizeof(buf));
          fname = fselect(buf, 0);
          if (!fname) return -1;/* let the user select a file */
          r = lvq_save(fname);  /* save reference vectors */
          if (r != 0) { error(r, fname); return -1; }
          return 0;

        case MI_QUIT:           /* File > Quit */
          DestroyWindow(hwnd);  /* destroy main window */
          return 0;             /* and abort the program */

        /* --- Menu Actions --- */
        case MI_INIT:           /* 'Actions > Initialize' */
          if (!table) { error(E_DATA); return 0; }
          lvq_init();           /* init. the reference vectors */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        case MI_START:          /* 'Actions > Start Training' */
          if (!table) { error(E_DATA); return 0; }
          if (!shades) return 0;
          if (!rvecs) lvq_init();
          if (!rvecs) return 0; /* initialize reference vectors */
          mn_enable(0);         /* disable certain menu items */
          running = 1;          /* set the running flag */
          train(hmain, WM_TIMER, 1, 0);
          return 0;             /* start the training */

        case MI_STOP:           /* 'Actions > Stop Training' */
          KillTimer(hmain, 1);  /* kill the timer */
          mn_enable(1);         /* enable all menu items */
          return running = 0;   /* clear the running flag */

        case MI_REDRAW:         /* 'Actions > Redraw' */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */

        /* --- Menu Settings --- */
        case MI_PARAMS:         /* 'Settings > Parameters...' */
          if (DialogBox(hinst, MKIRSC(DB_PARAMS), hwnd,
             (DLGPROC)dp_params) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* on ok redraw the window contents */

        case MI_ATTS:           /* 'Settings > Attributes...' */
          if (DialogBox(hinst, MKIRSC(DB_ATTS), hwnd,
             (DLGPROC)dp_atts) != 0)
            return 0;           /* on ok remove existing ref. vectors */
          lvq_clean();          /* and get new color shades */
          if (col_create() != 0) { error(E_COLOR); return -1; }
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* on ok redraw the window contents */

        case MI_COLORS:         /* 'Settings > Colors...' */
          if (DialogBox(hinst, MKIRSC(DB_COLORS), hwnd,
             (DLGPROC)dp_colors) != 0)
            return 0;           /* on ok get new color shades */
          if (col_create() != 0) { error(E_COLOR); return -1; }
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* on ok redraw the window contents */

        /* --- Menu Help --- */
        case MI_ABOUT:          /* 'Help > About...' */
          DialogBox(hinst, MKIRSC(DB_ABOUT), hwnd, (DLGPROC)dp_about);
          return 0;             /* show 'about' dialog box */
      } break;

    case WM_CLOSE:              /* window closed */
    case WM_DESTROY:            /* or destroyed */
      PostQuitMessage(0);       /* terminate program */
      return 0;
  }
  return DefWindowProc(hwnd, msg, wp, lp);
}  /* wp_main() */

/*----------------------------------------------------------------------
  Main Program
----------------------------------------------------------------------*/

int WINAPI WinMain (HINSTANCE hcurr, HINSTANCE hprev,
                    LPSTR cmdln, int cmd)
{                               /* --- main function */
  WNDCLASS wc;                  /* window class */
  RECT     rc;                  /* window rectangle */
  MSG      msg;                 /* message */
  char     *arg;                /* command line argument */
  char     *s;                  /* to traverse the command line */
  int      c;                   /* buffer for a character */

  /* --- create window class --- */
  hinst = hcurr;                /* note program instance handle */
  if (!hprev) {                 /* if first instance of this program */
    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = wp_main;
    wc.cbClsExtra    = wc.cbWndExtra = 0;
    wc.hInstance     = hinst;
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName  = MKIRSC(MN_MAIN);
    wc.lpszClassName = "WLVQClass";
    if (!RegisterClass(&wc))    /* register class of */
      return FALSE;             /* main window */
  }

  /* --- get brushes and colors --- */
  p_black  = GetStockObject(BLACK_PEN);
  p_white  = GetStockObject(WHITE_PEN);
  br_null  = GetStockObject(HOLLOW_BRUSH);
  br_white = GetStockObject(WHITE_BRUSH);
  br_black = GetStockObject(BLACK_BRUSH);
  br_grey  = CreateSolidBrush(RGB(160, 160, 160));

  /* --- create main window --- */
  rc.left = 30; rc.right  = 30 +INITXEXT -1;
  rc.top  = 30; rc.bottom = 30 +INITYEXT -1;
  AdjustWindowRect(&rc, WINSTYLE, 1);
  hmain = CreateWindow("WLVQClass", PRGNAME, WINSTYLE,
                       30, 30, rc.right-rc.left, rc.bottom-rc.top,
                       NULL, NULL, hinst, NULL);
  if (hmain == 0) return FALSE; /* create main window */

  /* --- load initial vectors and data --- */
  while (cmdln && *cmdln) {     /* if a command line is given */
    s = cmdln;                  /* traverse the comand line */
    while (*s == ' ') s++;      /* find start of file name */
    if (*s == '\"') { arg = s+1; c = '\"'; }
    else            { arg = s;   c = ' ';  }
    while ((*s != c) && (*s != 0)) s++;
    c = *s; *s++ = 0;           /* find end of file name */
    if (*arg && (tv_load(arg, 0.1) != 0)) {
      error(E_FREAD, arg); break; }
    if (col_create() != 0) { error(E_COLOR); break; }
    if (!c) break;              /* check for another argument */
    while (*s == ' ') s++;      /* find start of file name */
    if (*s == '\"') { arg = s+1; c = '\"'; }
    else            { arg = s;   c = ' ';  }
    while ((*s != c) && (*s != 0)) s++;
    *s = 0;                     /* find end of file name */
    if (*arg && (lvq_load(arg) != 0)) {
      error(E_FREAD, arg); break; }
    break;                      /* terminate dummy loop */
  }  /* while (cmdln && *cmdln) */

  /* --- show main window --- */
  ShowWindow(hmain, cmd);       /* open and */
  UpdateWindow(hmain);          /* update main window */
  hdc = GetDC(hmain);           /* get a drawing device context */

  /* --- dispatch messages --- */
  while (GetMessage(&msg, NULL, 0, 0)) {
    TranslateMessage(&msg);     /* get messages, translate */
    DispatchMessage (&msg);     /* and dispatch them */
  }

  /* --- clean up --- */
  ReleaseDC(hmain, hdc);        /* release the device context */
  lvq_clean();                  /* clean up LVQ variables and */
  tv_clean();                   /* table visualization variables */
  col_delete();                 /* delete colors/shades etc. */
  return (int)msg.wParam;
}  /* WinMain() */
